/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Jan-Michael Brummer                                  *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file menucallbacks.c
 * \brief contains the callbacks used in menus 
 */

#include <ffgtk.h>

#ifdef HAVE_APPINDICATOR
#include <libappindicator/app-indicator.h>
AppIndicator *psIndicator = NULL;
#endif

/** trayicon open state */
bool bOpen = FALSE;

/* Widget containing call monitor */
static GtkWidget *psMonitor = NULL;

/** main gtk loop */
extern GMainLoop *psLoop;

/**
 * \brief Destroy callback
 * \param psWidget widget
 * \param pData optional data pointer
 */

gboolean destroy( GtkWidget *psWidget, gpointer pData ) {
	g_main_quit( psLoop );
	return TRUE;
}

/**
 * \brief Set try icon state
 * \param bSet set state
 */
void setTrayOpen( bool bSet ) {
	bOpen = bSet;
}

/** 
 * \brief Open CallMonitor
 */

void openCallMonitor( void ) {

	if ( bOpen == false ) {
		psMonitor = CallMonitor();
		if ( psMonitor != NULL ) {
			bOpen = TRUE;
		}
	}	
}

/**
 * Close CallMonitor
 */

void closeCallMonitor( void ) {

	if ( bOpen == true ) {
		if ( psMonitor != NULL ) {
			gtk_widget_hide( GTK_WIDGET( psMonitor ) );
		}
		bOpen = FALSE;
	}	
}

/**
 * \brief Called on tray icon activated, opens or close CallMonitor
 */
void trayIconToggled( void ) {

	if ( bOpen == false ) {
		openCallMonitor();
	} else {
		closeCallMonitor();
	}
	
#ifndef HAVE_APPINDICATOR
	menuSetIconType( ICON_CALLOUT );
#else
	app_indicator_set_icon( psIndicator, "ffgtk" );
#endif
}

/**
 * \brief Open addressbook - callback
 */
void trayIconAddressBook( void ) {
	AddressBook( 0 );
}

/**
 * \brief Select active profile callback
 * \param psItem menu item pressed
 * \param pUserData name of profile
 */
void trayIconProfile( GtkMenuItem *psItem, gpointer pUserData ) {
	if ( gtk_check_menu_item_get_active( GTK_CHECK_MENU_ITEM( psItem ) ) == TRUE ) {
		const gchar *pnName = pUserData;
		struct sProfile *psProfile = findProfile( pnName );
		setActiveProfile( psProfile );
		saveProfiles();
	}
}

/**
 * \brief Dial number - callback
 * \param psWidget menu item widget
 * \param pnNumber dial number
 */
void trayIconDialNumber( GtkWidget *psWidget, gchar *pnNumber ) {
	struct sPerson *psPerson = NULL;

	if ( pnNumber != NULL ) {
		psPerson = findPersonByNumber( pnNumber );
	}

	dialNumberDialog( psPerson, pnNumber );
}

/**
 * \brief Call router reconnect - callback
 */
void trayIconReconnect( void ) {
	routerReconnect( getActiveProfile() );
}

/**
 * \brief Copy external ip to clipboard - callback
 */
void trayIconIpToClipboard( void ) {
	gchar *pnIp = routerGetIp( getActiveProfile() );

	if ( pnIp != NULL ) {
		gtk_clipboard_set_text( gtk_clipboard_get( GDK_NONE ), pnIp, -1 );
		g_free( pnIp );
	}
}

/**
 * \brief Mute sounds - callback
 * \param psItem menu item widget
 * \param pUserData UNUSED
 */
void trayIconMute( GtkCheckMenuItem *psItem, gpointer pUserData ) {
	gboolean bMute = callMonitorGetMute( getActiveProfile() );

	gtk_check_menu_item_set_active( psItem, !bMute );
	callMonitorSetMute( getActiveProfile(), !bMute );
}

/**
 * \brief Set diversity state - callback
 * \param psItem menu item widget
 * \param psDiversity diversity structure
 */
void trayIconSetDiversity( GtkMenuItem *psItem, struct sDiversity *psDiversity ) {
	psDiversity -> bSet = !psDiversity -> bSet;

	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM( psItem ), psDiversity -> bSet );
	routerSetDiversity( getActiveProfile(), psDiversity -> nIndex, psDiversity -> bSet );
}

#if !defined(G_OS_WIN32) && !defined(__MACOSX__)
/**
 * \brief Send SMS - callback
 * \param psItem menu item widget
 * \param pUserData UNUSED
 */
void trayIconSendSms( GtkMenuItem *psItem, gpointer pUserData ) {
	sendSmsDialog();
}
#endif

/**
 * \brief Open help - callback
 * \param psItem menu item widget
 * \param psDiversity diversity structure
 */
void trayIconHelp( GtkMenuItem *psItem, struct sDiversity *psDiversity ) {
	osExecute( "http://www.tabos.org/forum" );
}

void menuBarSetActiveProfile( GtkAction *start, GtkAction *current ) {	
	const gchar *pName = gtk_action_get_label( GTK_ACTION( current ) );

	struct sProfile *psProfile = findProfile( pName );
	setActiveProfile( psProfile );
	
}	

void menuBarDialNumber( GtkAction *dialAction ) {
	const gchar *pAction  = gtk_action_get_name( GTK_ACTION( dialAction) );
	struct sPerson *psPerson = NULL;

	gchar *pnNumber = g_strrstr( pAction, "Action") + 6;
	
        if ( pnNumber != NULL ) {
                psPerson = findPersonByNumber( pnNumber );
        }

        dialNumberDialog( psPerson, pnNumber );
}
